<?php

class CM_SessionTest extends CMTest_TestCase {

    public function testConstructor() {
        $session = new CM_Session();
        $this->assertTrue(true);

        try {
            new CM_Session('nonexistent');
            $this->fail('Can instantiate nonexistent session.');
        } catch (CM_Exception_UnknownSessionId $ex) {
            $this->assertTrue(true);
        }
    }

    public function testSetGetDelete() {
        $session = new CM_Session();
        $this->assertNull($session->get('foo'));

        $session->set('foo', 'bar');
        $this->assertSame('bar', $session->get('foo'));

        $session->delete('foo');
        $this->assertNull($session->get('foo'));

        $session->set('bar', array('foo', 'bar'));
        $this->assertEquals(array('foo', 'bar'), $session->get('bar'));
    }

    public function testPersistence() {
        $session = new CM_Session();
        $sessionId = $session->getId();
        unset($session);
        try {
            new CM_Session($sessionId);
            $this->fail('Empty Session stored in db.');
        } catch (CM_Exception_UnknownSessionId $ex) {
            $this->assertTrue(true);
        }
        $session = new CM_Session();
        $sessionId = $session->getId();
        $session->set('foo', 'bar');
        unset($session);
        $session = new CM_Session($sessionId);

        $session->set('bar', array('foo', 'bar'));
        $session->set('foobar', 'foobar');
        $expiration = $session->getExpiration();
        $sessionId = $session->getId();
        CMTest_TH::timeForward(10);
        unset($session);

        try {
            $session = new CM_Session($sessionId);
            $this->assertTrue(true);
        } catch (CM_Exception_UnknownSessionId $ex) {
            $this->fail('Session not persistent.');
        }
        $this->assertEquals('bar', $session->get('foo'));
        $this->assertEquals(array('foo', 'bar'), $session->get('bar'));
        $this->assertEquals($expiration + 10, $session->getExpiration(), null, 1);

        //test that session is only persisted when data changed
        CM_Db_Db::update('cm_session', array('data' => serialize(array('foo' => 'foo', 'foobar' => 'foobar'))),
            array('sessionId' => $session->getId()));
        CMTest_TH::clearCache();
        unset($session);
        $session = new CM_Session($sessionId);
        $this->assertEquals('foo', $session->get('foo'));

        $session->delete('foobar');
        unset($session);
        $session = new CM_Session($sessionId);
        $this->assertNull($session->get('foobar'));

        //caching
        $session->set('foo', 'foo');
        $sessionId = $session->getId();
        unset($session);
        $session = new CM_Session($sessionId);
        unset($session);

        CM_Db_Db::update('cm_session', array('data' => serialize(array('foo' => 'bar'))), array('sessionId' => $sessionId));
        $session = new CM_Session($sessionId);
        $this->assertEquals('foo', $session->get('foo'));

        $session->delete('foo');
        $this->assertTrue($session->isEmpty());
        unset($session);
        try {
            $session = new CM_Session($sessionId);
            $this->fail('Session not deleted.');
        } catch (CM_Exception_UnknownSessionId $ex) {
            $this->assertTrue(true);
        }
    }

    public function testRegenerateId() {
        $session = new CM_Session();
        $session->set('foo', 'bar');
        $sessionId = $session->getId();
        unset($session);
        $session = new CM_Session($sessionId);
        $oldSessionId = $session->getId();
        $session->regenerateId();
        $newSessionId = $session->getId();
        unset($session);
        try {
            new CM_Session($oldSessionId);
            $this->fail('Db entry not updated.');
        } catch (CM_Exception_UnknownSessionId $ex) {
            $this->assertTrue(true);
        }
        try {
            $session = new CM_Session($newSessionId);
            $this->assertTrue(true);
        } catch (CM_Exception_Nonexistent $ex) {
            $this->fail('Db entry not updated.');
        }
        $this->assertEquals('bar', $session->get('foo'));
    }

    public function testGc() {
        $session = new CM_Session();
        $sessionId = $session->getId();
        unset($session);
        CMTest_TH::timeForward(4000);
        CM_Session::deleteExpired();
        try {
            new CM_Session($sessionId);
            $this->fail('Expired Session was not deleted.');
        } catch (CM_Exception_UnknownSessionId $ex) {
            $this->assertTrue(true);
        }
    }

    public function testSetUser() {
        $user = CM_Model_User::createStatic();
        $session = new CM_Session();

        $session->setUser($user);
        $this->assertEquals($user, $session->getUser(true));
        $this->assertTrue($session->getUser(true)->getOnline());
    }

    public function testSetUserWithRequest() {
        $user = CM_Model_User::createStatic();
        $request = new CM_Http_Request_Get('/', ['user-agent' => 'Chrome']);
        $session = new CM_Session(null, $request);

        $session->setUser($user);
        $userAgentList = $user->getUseragents();
        $this->assertSame([['useragent' => 'Chrome', 'createStamp' => time()]], $userAgentList->getItems());
    }

    public function testDeleteUser() {
        $session = new CM_Session();
        $session->setUser(CM_Model_User::createStatic());
        $user = $session->getUser(true);

        $session->deleteUser();
        $this->assertNull($session->getUser());
        $user->_change();
        $this->assertFalse($user->getOnline());
    }

    public function testLogoutUser() {
        $session = new CM_Session();
        $session->setUser(CM_Model_User::createStatic());
        $session->setLifetime(123);
        $this->assertNotNull($session->getUser());
        $this->assertTrue($session->hasLifetime());

        $session->logoutUser();
        $this->assertNull($session->getUser());
        $this->assertFalse($session->hasLifetime());
    }

    public function testGetViewer() {
        $session = new CM_Session();
        $this->assertNull($session->getUser());
        try {
            $session->getUser(true);
            $this->fail('Should throw exception');
        } catch (CM_Exception_AuthRequired $ex) {
            $this->assertTrue(true);
        }

        /** @var CM_Model_User $user */
        $user = CM_Model_User::createStatic();
        $session->setUser($user);
        $this->assertEquals($user, $session->getUser(true));
    }

    public function testStart() {
        /** @var CM_Model_User $user */
        $user = CM_Model_User::createStatic();
        $user->updateLatestActivityThrottled();
        $activityStamp1 = time();
        $session = new CM_Session();
        $session->setUser($user);
        $sessionId = $session->getId();
        unset($session);
        $session = new CM_Session($sessionId);
        $this->assertEquals($activityStamp1, $session->getUser(true)->getLatestActivity(), null, 1);

        CMTest_TH::timeForward(CM_Model_User::ACTIVITY_EXPIRATION / 2);
        $session = new CM_Session($sessionId);
        $session->start();
        $this->assertEquals($activityStamp1, $session->getUser(true)->getLatestActivity(), null, 1);

        CM_Db_Db::update('cm_session', array('data' => serialize(array('userId' => $user->getId(), 'foo' => 'bar'))));
        unset($session);
        CMTest_TH::clearCache();

        CMTest_TH::timeForward(CM_Model_User::ACTIVITY_EXPIRATION / 2 + 1);
        $activityStamp2 = time();
        $session = new CM_Session($sessionId);
        $session->start();
        $this->assertEquals($activityStamp2, $session->getUser(true)->getLatestActivity(), null, 1);
        CMTest_TH::timeForward($session->getLifetime() / 2);
        $session->start();

        $this->assertEquals('bar', $session->get('foo'));
        CM_Db_Db::update('cm_session', array('data' => serialize(array('userId' => $user->getId(), 'foo' => 'foo'))));
        unset($session);
        CMTest_TH::clearCache();

        $session = new CM_Session($sessionId);
        $this->assertEquals('bar', $session->get('foo'));
    }

    public function testStartUpdateUserAgents() {
        $userClassMock = $this->mockClass('CM_Model_User');
        /** @var CM_Model_User|\Mocka\AbstractClassTrait $user */
        $user = $userClassMock->newInstance([CMTest_TH::createUser()->getId()]);

        $sessionClassMock = $this->mockClass('CM_Session');
        $sessionClassMock->mockMethod('getUser')->set($user);

        /** @var CM_Paging_Useragent_User|\Mocka\AbstractClassTrait $userAgents */
        $userAgents = $this->mockClass('CM_Paging_Useragent_User')->newInstance([$user]);
        $uaAddMethod = $userAgents->mockMethod('addFromRequest');
        $uaAddMethod
            ->at(0, function (CM_Http_Request_Abstract $request) {
                $this->assertSame('Chrome', $request->getUserAgent());
            })
            ->at(1, function (CM_Http_Request_Abstract $request) {
                $this->assertSame('Chrome', $request->getUserAgent());
            })
            ->at(2, function (CM_Http_Request_Abstract $request) {
                $this->assertSame('Chrome Foo', $request->getUserAgent());
            });

        $userClassMock->mockMethod('getUseragents')->set($userAgents);

        // tests

        $request = new CM_Http_Request_Get('/', ['user-agent' => 'Chrome']);
        /** @var CM_Session|\Mocka\AbstractClassTrait $session */
        $session = $sessionClassMock->newInstance([null, $request]);

        $this->assertSame(0, $uaAddMethod->getCallCount());
        $session->setUser($user);
        $this->assertSame(1, $uaAddMethod->getCallCount());
        $session->start();
        $this->assertSame(1, $uaAddMethod->getCallCount());

        $sessionId = $session->getId();
        unset($session);
        $request = new CM_Http_Request_Get('/', ['user-agent' => 'Chrome']);
        /** @var CM_Session|\Mocka\AbstractClassTrait $session */
        $session = $sessionClassMock->newInstance([$sessionId, $request]);

        $this->assertSame(1, $uaAddMethod->getCallCount());
        $session->start();
        $this->assertSame(1, $uaAddMethod->getCallCount());

        unset($session);
        $request = new CM_Http_Request_Get('/', ['user-agent' => 'Chrome']);
        /** @var CM_Session|\Mocka\AbstractClassTrait $session */
        $session = $sessionClassMock->newInstance([$sessionId, $request]);
        $user->setOnline(false);

        $this->assertSame(1, $uaAddMethod->getCallCount());
        $session->start();
        $this->assertSame(2, $uaAddMethod->getCallCount());

        unset($session);
        $request = new CM_Http_Request_Get('/', ['user-agent' => 'Chrome Foo']);
        /** @var CM_Session|\Mocka\AbstractClassTrait $session */
        $session = $sessionClassMock->newInstance([$sessionId, $request]);
        $user->setOnline(false);

        $this->assertSame(2, $uaAddMethod->getCallCount());
        $session->start();
        $this->assertSame(3, $uaAddMethod->getCallCount());
    }

    public function testExpiration() {
        $session = new CM_Session();
        $session->set('foo', 'bar');
        $sessionId = $session->getId();
        unset($session);
        $session = new CM_Session($sessionId);
        $this->assertEquals(time() + CM_Session::LIFETIME_DEFAULT, $session->getExpiration(), null, 1);

        $session->setLifetime(10 * 86400);
        unset($session);
        $session = new CM_Session($sessionId);
        $this->assertEquals(time() + 10 * 86400, $session->getExpiration(), null, 1);

        $session->setLifetime();
        $sessionId = $session->getId();
        unset($session);
        $session = new CM_Session($sessionId);
        $this->assertEquals(time() + CM_Session::LIFETIME_DEFAULT, $session->getExpiration(), null, 1);
    }

    public function testFindById() {
        $session = new CM_Session();
        $session->set('foo', 'bar');
        $session->write();
        $this->assertEquals($session, CM_Session::findById($session->getId()));

        $this->assertNull(CM_Session::findById('foo'));
        $this->assertNull(CM_Session::findById(''));
    }
}
